@@ -3,17 +3,21 @@ from typing import Optional, Union, List, Set, Tuple, Any, Callable, Type
 from abc import abstractmethod, ABCMeta
 import json
 import warnings
-import os
 from collections import OrderedDict
 import importlib
+import os
 
 import spode.core.const as const
-# import const
 
-model_json_name = 'model.json'
-all_model_names = ['WaveGuide', 'PhaseShift', 'DirectionalCoupler', 'TBUm']
 
-__all__ = ['GeneralModel'] + all_model_names + ['reset_model_json', 'register_model']
+# the model json file. It will be used as reference when parsing the circuit from a netlist file
+_model_json_name = os.path.join(os.path.dirname(__file__), 'model.json')
+
+# the built-in models defined in this file.
+_all_class_names = ['WaveGuide', 'PhaseShift', 'DirectionalCoupler', 'TBUm', 'TBUo', 'TBUt']
+
+# the models and methods which are visible to users
+__all__ = ['GeneralModel'] + _all_class_names + ['reset_model_json', 'register_model']
 
 
 def _neff(neff0: float, ng0: Optional[float] = None, wl0: Optional[float] = None) -> Callable:
@@ -51,7 +55,7 @@ class GeneralModel(object, metaclass=ABCMeta):
         _required_attr: mandatory attributes of the circuit model. Their values must be provided.
         _optional_attr: optional attributes of the circuit model. If omitted, they will be set to default values.
         _differential_attr: the attributes that the derivatives could be evaluated w.r.t.
-        _num_port: the number of ports at the left end and right end.
+        _num_port: a list of two elements, the number of ports at the left end and right end.
 
     Instance Variables:
         params: the circuit model parameters and the associated values in the format of key-value pair.
@@ -83,7 +87,7 @@ class GeneralModel(object, metaclass=ABCMeta):
                 self.params[attr] = kwargs[attr]
             else:
                 raise RuntimeError("The value of attribute '%s' is not provided when initializing model '%s'."
-                                   %(attr, self.__class__.__name__))
+                                   % (attr, self.__class__.__name__))
 
         for attr, default_value in self._optional_attr.items():
             self.params[attr] = kwargs[attr] if attr in kwargs.keys() else default_value
@@ -152,6 +156,7 @@ class GeneralModel(object, metaclass=ABCMeta):
 
         info_dict = OrderedDict()
         info_dict['model_name'] = cls._name
+        info_dict['class_name'] = cls.__name__
         info_dict['required_attr'] = [attr for attr in cls._required_attr]
         info_dict['optional_attr'] = [attr for attr in cls._optional_attr]
         info_dict['differential_attr'] = cls._differential_attr
@@ -194,7 +199,7 @@ class WaveGuide(GeneralModel):
 
         # linearly approximate the effective index neff
         neff_func = _neff(self.params['neff'], self.params['ng'], self.params['wl'])
-        beta = omega * neff_func(omega) / const.FreeLightSpeed  # propagation constant
+        beta = (omega * neff_func(omega) / const.FreeLightSpeed).reshape(-1,1,1)  # propagation constant
 
         # define and reshape the Smatrix to the correct shape (len(omega), self._num_port[0], self._num_port[1])
         # This model doesn't use the variable direction.
@@ -202,14 +207,12 @@ class WaveGuide(GeneralModel):
 
         if deri_vari is not None:
             grad = {}
-            if deri_vari.lower() == 'all':
-                deri_vari = self._differential_attr
 
             if 'L' in deri_vari:
                 grad['L'] = 1.j * beta * Smatrix
 
             if 'alpha' in deri_vari:
-                grad['alpha'] = Smatrix / self.params['alpha']
+                grad['alpha'] = 1.0 / self.params['alpha'] * Smatrix
 
             return Smatrix, grad
         else:
@@ -244,8 +247,6 @@ class PhaseShift(GeneralModel):
 
         if deri_vari is not None:
             grad = {}
-            if deri_vari.lower() == 'all':
-                deri_vari = self._differential_attr
 
             if 'ps' in deri_vari:
                 grad['ps'] = 1.j * Smatrix
@@ -288,8 +289,6 @@ class DirectionalCoupler(GeneralModel):
 
         if deri_vari is not None:
             grad = {}
-            if deri_vari.lower() == 'all':
-                deri_vari = self._differential_attr
 
             if 'cp_angle' in deri_vari:
                 grad_cp_angle = self.params['alpha'] * np.array([[-np.sin(theta), 1.j * np.cos(theta)],
@@ -307,10 +306,11 @@ class DirectionalCoupler(GeneralModel):
 
 class TBUm(GeneralModel):
     """The class definition of a tunable basic unit (tbu) model with all phase shifts in the middle."""
+    #TODO: add paper refernce, who uses this model.
 
     _name = 'tbum'
     _required_attr = ['theta', 'phi', 'L', 'neff']
-    _optional_attr = {'ng': 0, 'wl': None, 'alpha': 1.0, 'cp_left': 0.5 * np.pi, 'cp_right': 0.5 * np.pi}
+    _optional_attr = {'ng': 0, 'wl': None, 'alpha': 1.0, 'cp_left': 0.25 * np.pi, 'cp_right': 0.25 * np.pi}
     _differential_attr = ['theta', 'phi', 'L', 'cp_left', 'cp_right']
     _num_port = [2, 2]
 
@@ -354,19 +354,17 @@ class TBUm(GeneralModel):
             S = S3 @ S2 @ S1
             gradS_theta = S3 @ np.array([[1.j * np.exp(1.j * self.params['theta']), 0.], [0., 0.]]) @ S1
             gradS_phi = S3 @ np.array([[0., 0.], [0., 1.j * np.exp(1.j * self.params['phi'])]]) @ S1
-            # TODO: the warnign message printed is bit messy; could be improved.
+            # TODO: the warning message printed is bit messy; could be improved.
             warnings.warn("The variable 'direction' should be either 'l2r' or 'r2l' when building S matrix. "
                           "Automatically use 'direction'='l2r'. This might lead to error, if the "
                           "two directional couplers in the {} are not identical.".format(self._name))
 
-        S_wg = np.exp(1.j * beta * self.params['L']) * self.params['alpha']
+        S_wg = (np.exp(1.j * beta * self.params['L']) * self.params['alpha']).reshape(-1, 1, 1)
 
-        Smatrix = S_wg.reshape(-1, 1, 1) * np.expand_dims(S, axis=0).repeat(len(omega), axis=0)
+        Smatrix = S_wg * np.expand_dims(S, axis=0).repeat(len(omega), axis=0)
 
         if deri_vari is not None:
             grad = {}
-            if deri_vari.lower() == 'all':
-                deri_vari = self._differential_attr
 
             if 'theta' in deri_vari:
                 grad['theta'] = S_wg * gradS_theta
@@ -384,50 +382,224 @@ class TBUm(GeneralModel):
         else:
             return Smatrix
 
+class TBUo(GeneralModel):
+    """The class definition of a tunable basic unit (tbu) model with phase shifts in the middle and right,
+    but in one row.
 
-def reset_model_json(model_name_list: List = all_model_names) -> bool:
+    This model has been used by some papers, such as S. Bandyopadhyay et al., 'Hardware error correction for
+    programmable photonics', Optica.
+
+    """
+
+    _name = 'tbuo'
+    _required_attr = ['theta', 'phi', 'L', 'neff']
+    _optional_attr = {'ng': 0, 'wl': None, 'alpha': 1.0, 'cp_left': 0.25 * np.pi, 'cp_right': 0.25 * np.pi}
+    _differential_attr = ['theta', 'phi', 'L', 'cp_left', 'cp_right']
+    _num_port = [2, 2]
+
+    def __init__(self, **kwargs):
+        super().__init__(**kwargs)
+
+    def get_smatrix(self,
+                    omega: Union[list, np.ndarray],
+                    direction: Optional[str] = None,
+                    deri_vari: Optional[Union[str, List[str], Set[str]]] = None) -> \
+            Union[Tuple[np.ndarray, dict], np.ndarray]:
+
+        # linearly approximate the effective index neff
+        neff_func = _neff(self.params['neff'], self.params['ng'], self.params['wl'])
+        beta = omega * neff_func(omega) / const.FreeLightSpeed  # propagation constant
+
+        # define and reshape the Smatrix to the correct shape (len(omega), self._num_port[0], self._num_port[1])
+
+        S1 = np.zeros((2, 2), dtype=np.complex_)
+        S1[0, 0] = S1[1, 1] = 1.0 * np.cos(self.params['cp_left'])
+        S1[0, 1] = S1[1, 0] = 1.j * np.sin(self.params['cp_left'])
+
+        S2 = np.zeros((2, 2), dtype=np.complex_)
+        S2[0, 0] = np.exp(1.j * self.params['theta'])
+        S2[1, 1] = 1.0
+
+        S3 = np.zeros((2, 2), dtype=np.complex_)
+        S3[0, 0] = S3[1, 1] = 1.0 * np.cos(self.params['cp_right'])
+        S3[0, 1] = S3[1, 0] = 1.j * np.sin(self.params['cp_right'])
+
+        S4 = np.zeros((2, 2), dtype=np.complex_)
+        S4[0, 0] = np.exp(1.j * self.params['phi'])
+        S4[1, 1] = 1.0
+
+        # If the left and right directional couplers are not identical, then the S matrix will be different.
+        if direction == 'r2l':
+            S = S1 @ S2 @ S3 @ S4
+            gradS_theta = S1 @ np.array([[1.j * np.exp(1.j * self.params['theta']), 0.], [0., 0.]]) @ S3 @ S4
+            gradS_phi = S1 @ S2 @ S3 @ np.array([[1.j * np.exp(1.j * self.params['phi']), 0.], [0., 0.]])
+        elif direction == 'l2r':
+            S = S4 @ S3 @ S2 @ S1
+            gradS_theta = S4 @ S3 @ np.array([[1.j * np.exp(1.j * self.params['theta']), 0.], [0., 0.]]) @ S1
+            gradS_phi = np.array([[1.j * np.exp(1.j * self.params['phi']), 0.], [0., 0.]]) @ S3 @ S2 @ S1
+        else:
+            S = S4 @ S3 @ S2 @ S1
+            gradS_theta = S4 @ S3 @ np.array([[1.j * np.exp(1.j * self.params['theta']), 0.], [0., 0.]]) @ S1
+            gradS_phi = np.array([[1.j * np.exp(1.j * self.params['phi']), 0.], [0., 0.]]) @ S3 @ S2 @ S1
+            warnings.warn("The variable 'direction' should be either 'l2r' or 'r2l' when building S matrix. "
+                          "Automatically use 'direction'='l2r'. This might lead to error, if the "
+                          "two directional couplers in the {} are not identical.".format(self._name))
+
+        S_wg = (np.exp(1.j * beta * self.params['L']) * self.params['alpha']).reshape(-1, 1, 1)
+
+        Smatrix = S_wg * np.expand_dims(S, axis=0).repeat(len(omega), axis=0)
+
+        if deri_vari is not None:
+            grad = {}
+
+            if 'theta' in deri_vari:
+                grad['theta'] = S_wg * gradS_theta
+
+            if 'phi' in deri_vari:
+                grad['phi'] = S_wg * gradS_phi
+
+            if 'alpha' in deri_vari:
+                grad['alpha'] = Smatrix / self.params['alpha']
+
+            if 'L' in deri_vari:
+                grad['L'] = Smatrix * 1.j * beta
+
+            return Smatrix, grad
+        else:
+            return Smatrix
+
+class TBUt(GeneralModel):
+    """The class definition of a tunable basic unit (tbu) model with phase shifts in the middle and right,
+    but in two rows."""
+
+    # TODO: add paper reference, who uses this model.
+
+    _name = 'tbut'
+    _required_attr = ['theta', 'phi', 'L', 'neff']
+    _optional_attr = {'ng': 0, 'wl': None, 'alpha': 1.0, 'cp_left': 0.25 * np.pi, 'cp_right': 0.25 * np.pi}
+    _differential_attr = ['theta', 'phi', 'L', 'cp_left', 'cp_right']
+    _num_port = [2, 2]
+
+    def __init__(self, **kwargs):
+        super().__init__(**kwargs)
+
+    def get_smatrix(self,
+                    omega: Union[list, np.ndarray],
+                    direction: Optional[str] = None,
+                    deri_vari: Optional[Union[str, List[str], Set[str]]] = None) -> \
+            Union[Tuple[np.ndarray, dict], np.ndarray]:
+
+        # linearly approximate the effective index neff
+        neff_func = _neff(self.params['neff'], self.params['ng'], self.params['wl'])
+        beta = omega * neff_func(omega) / const.FreeLightSpeed  # propagation constant
+
+        # define and reshape the Smatrix to the correct shape (len(omega), self._num_port[0], self._num_port[1])
+
+        S1 = np.zeros((2, 2), dtype=np.complex_)
+        S1[0, 0] = S1[1, 1] = 1.0 * np.cos(self.params['cp_left'])
+        S1[0, 1] = S1[1, 0] = 1.j * np.sin(self.params['cp_left'])
+
+        S2 = np.zeros((2, 2), dtype=np.complex_)
+        S2[0, 0] = np.exp(1.j * self.params['theta'])
+        S2[1, 1] = 1.0
+
+        S3 = np.zeros((2, 2), dtype=np.complex_)
+        S3[0, 0] = S3[1, 1] = 1.0 * np.cos(self.params['cp_right'])
+        S3[0, 1] = S3[1, 0] = 1.j * np.sin(self.params['cp_right'])
+
+        S4 = np.zeros((2, 2), dtype=np.complex_)
+        S4[0, 0] = 1.0
+        S4[1, 1] = np.exp(1.j * self.params['phi'])
+
+        # If the left and right directional couplers are not identical, then the S matrix will be different.
+        if direction == 'r2l':
+            S = S1 @ S2 @ S3 @ S4
+            gradS_theta = S1 @ np.array([[1.j * np.exp(1.j * self.params['theta']), 0.], [0., 0.]]) @ S3 @ S4
+            gradS_phi = S1 @ S2 @ S3 @ np.array([[0., 0.], [0., 1.j * np.exp(1.j * self.params['phi'])]])
+        elif direction == 'l2r':
+            S = S4 @ S3 @ S2 @ S1
+            gradS_theta = S4 @ S3 @ np.array([[1.j * np.exp(1.j * self.params['theta']), 0.], [0., 0.]]) @ S1
+            gradS_phi = np.array([[0., 0.], [0., 1.j * np.exp(1.j * self.params['phi'])]]) @ S3 @ S2 @ S1
+        else:
+            S = S4 @ S3 @ S2 @ S1
+            gradS_theta = S4 @ S3 @ np.array([[1.j * np.exp(1.j * self.params['theta']), 0.], [0., 0.]]) @ S1
+            gradS_phi = np.array([[0., 0.], [0., 1.j * np.exp(1.j * self.params['phi'])]]) @ S3 @ S2 @ S1
+            warnings.warn("The variable 'direction' should be either 'l2r' or 'r2l' when building S matrix. "
+                          "Automatically use 'direction'='l2r'. This might lead to error, if the "
+                          "two directional couplers in the {} are not identical.".format(self._name))
+
+        S_wg = (np.exp(1.j * beta * self.params['L']) * self.params['alpha']).reshape(-1, 1, 1)
+
+        Smatrix = S_wg * np.expand_dims(S, axis=0).repeat(len(omega), axis=0)
+
+        if deri_vari is not None:
+            grad = {}
+
+            if 'theta' in deri_vari:
+                grad['theta'] = S_wg * gradS_theta
+
+            if 'phi' in deri_vari:
+                grad['phi'] = S_wg * gradS_phi
+
+            if 'alpha' in deri_vari:
+                grad['alpha'] = Smatrix / self.params['alpha']
+
+            if 'L' in deri_vari:
+                grad['L'] = Smatrix * 1.j * beta
+
+            return Smatrix, grad
+        else:
+            return Smatrix
+
+def reset_model_json() -> bool:
+    """Reset the model json file to only reflect the built-in models defined in this file."""
     info_list = []
-    module = importlib.import_module("model")
-    for model_name in model_name_list:
+    module = importlib.import_module("spode.core.model")
+    for model_name in _all_class_names:
         class_ = getattr(module, model_name)
         info_list.append(class_._collect_info())
 
-    with open(model_json_name, 'w') as f:
+    with open(_model_json_name, 'w') as f:
         json.dump(info_list, f, indent=4, separators=(',', ':'))
 
-    print("Model json file has been reset to default successfully.")
+    print("model.json file has been reset to default successfully.")
 
     return True
 
 
 def register_model(model: Type[GeneralModel]) -> bool:
-    # check if the model class name or the _name is used by other already defined models
-    module = importlib.import_module("model")
-    for defined_model in all_model_names:
+    """Register a user-defined model to the model.json file.
+
+    Since the provided netlist will be parsed according to the model.json file. This function will correctly register
+    the user-defined circuit model, so that during parsing, it could be recognized.
+
+    :param model: The name of the class (e.g., WaveGuide, TBUm).
+    """
+    # check if the class name or the _name is used by other already defined models
+    module = importlib.import_module("spode.core.model")
+    for defined_model in _all_class_names:
         defined_class = getattr(module, defined_model)
         if model.__name__.lower() == defined_class.__name__.lower():
-            raise RuntimeError("The class name %s is already used" %defined_class.__name__)
+            raise RuntimeError("The class name %s is already used" % defined_class.__name__)
         if model._name.lower() == defined_class._name.lower():
             raise RuntimeError("The model name %s is already used" % defined_class._name)
 
-    with open(model_json_name, 'r') as f:
+    with open(_model_json_name, 'r') as f:
         info = json.load(f)
 
     info.append(model._collect_info())
 
-    with open(model_json_name, 'w') as f:
+    with open(_model_json_name, 'w') as f:
         json.dump(info, f, indent=4, separators=(',', ':'))
 
-    print("The new model %s has been added to the simulator successfully." %(model.__name__))
+    print("The new model '%s' has been added successfully." % (model.__name__))
     return True
 
 
 if __name__ == '__main__':
-
     print("hello world. This is the wrong file to run. "
           "If you insist, it will do a testing run.")
 
-
     omega = np.linspace(0, 10, 10)
     wg = WaveGuide(L=3., neff=2.)
     wg.print_allattr()
@@ -455,22 +627,25 @@ if __name__ == '__main__':
 
     reset_model_json()
 
-    class test(GeneralModel):
 
+    class test(GeneralModel):
         _name = 'testuser'
         _required_attr = ['a1', 'a2']
         _optional_attr = ['a3', 'a4']
         _differential_attr = ['a4']
-        _num_port = [2,2]
+        _num_port = [2, 2]
 
         def __init__(self):
             super().__init__()
+
         def get_smatrix(self,
-                omega: Union[list, np.ndarray],
-                direction: Optional[str] = None,
-                deri_vari: Optional[Union[str, List[str], Set[str]]] = None) -> \
-        Union[Tuple[np.ndarray, dict], np.ndarray]:
-            return 1
+                        omega: Union[list, np.ndarray],
+                        direction: Optional[str] = None,
+                        deri_vari: Optional[Union[str, List[str], Set[str]]] = None) -> \
+                Union[Tuple[np.ndarray, dict], np.ndarray]:
+            return np.array([1])
+
 
     register_model(test)
 
+    reset_model_json()
